import subprocess

import pytest
from keyring.errors import PasswordDeleteError

from onepassword_keyring import OnePasswordBackend, _onepassword_cli_installed, _user_is_authenticated, _item_exists, \
    sign_in, cli_exec


def mock_auth(mocker):
    mocker.patch('onepassword_keyring._auth', return_value=True)


def mock_item_exists(mocker, exists=True):
    mocker.patch('onepassword_keyring._item_exists', return_value=exists)


def mock_cli_exec(mocker, stdout="", stderr=""):
    mocker.patch('onepassword_keyring.cli_exec', return_value=(stdout, stderr))


def test_cli_exec(mocker):
    mocker.patch('subprocess.run', return_value=subprocess.CompletedProcess(args=[], returncode=0, stdout='output', stderr=''))
    stdout, stderr = cli_exec("some_command")
    assert stdout == 'output'
    assert stderr == ''


@pytest.mark.parametrize("path, expected", [(None, False), ("yay", True)])
def test_onepassword_cli_installed(mocker, path, expected):
    mocker.patch("shutil.which", return_value=path)
    assert _onepassword_cli_installed() == expected


def test_onepassword_not_installed(mocker):
    mocker.patch('shutil.which', return_value=None)
    with pytest.raises(RuntimeError):
        OnePasswordBackend.priority()


def test_user_not_authenticated(mocker):
    mock_cli_exec(mocker, stdout="", stderr="Not authenticated")
    assert not _user_is_authenticated()


def test_sign_in_successful(mocker):
    mock_cli_exec(mocker, stdout='Signed in successfully', stderr='')
    assert sign_in() is True


def test_sign_in_failed(mocker):
    mock_cli_exec(mocker, stdout='', stderr='Failed to sign in')
    with pytest.raises(RuntimeError):
        sign_in()


def test_item_exists(mocker):
    mock_cli_exec(mocker, stdout="user123", stderr="")
    assert _item_exists("some_service", "user123")


def test_item_not_exists(mocker):
    mock_cli_exec(mocker, stdout="user321", stderr="")
    assert not _item_exists("some_service", "user123")


def test_get_password(mocker):
    mock_auth(mocker)
    mock_item_exists(mocker)
    mock_cli_exec(mocker, stdout="user123,password123")

    backend = OnePasswordBackend()
    password = backend.get_password("some_service", "user123")

    assert password == "password123"


def test_get_password_error(mocker):
    mock_auth(mocker)
    mock_item_exists(mocker)
    mock_cli_exec(mocker, stdout="", stderr="Some error")

    backend = OnePasswordBackend()
    with pytest.raises(RuntimeError):
        backend.get_password("some_service", "user123")


def test_set_password(mocker):
    mock_auth(mocker)
    mock_item_exists(mocker, exists=False)
    mock_cli_exec(mocker)

    backend = OnePasswordBackend()
    backend.set_password("some_service", "user123", "password123")


def test_set_password_item_exists(mocker):
    mock_auth(mocker)
    mock_item_exists(mocker, exists=True)
    mock_cli_exec(mocker)

    backend = OnePasswordBackend()
    backend.set_password("some_service", "user123", "new_password123")


def test_delete_password(mocker):
    mock_auth(mocker)
    mock_item_exists(mocker)
    mock_cli_exec(mocker)

    backend = OnePasswordBackend()
    backend.delete_password("some_service", "user123")


def test_delete_password_item_not_exists(mocker):
    mock_auth(mocker)
    mock_item_exists(mocker, exists=False)
    mock_cli_exec(mocker)

    backend = OnePasswordBackend()
    with pytest.raises(PasswordDeleteError):
        backend.delete_password("some_service", "user123")


def test_get_otp(mocker):
    mock_auth(mocker)
    mock_item_exists(mocker)
    mock_cli_exec(mocker, stdout="otp_123")

    backend = OnePasswordBackend()
    otp = backend.get_otp("some_service", "user123")

    assert otp == "otp_123"


def test_get_item(mocker):
    mock_auth(mocker)
    mock_item_exists(mocker)
    mock_cli_exec(mocker, stdout="item_info")

    backend = OnePasswordBackend()
    item = backend.get_item("some_service")

    assert item == "item_info"



import shutil
import subprocess
import sys

from keyring import backend
from jaraco.classes import properties
from keyring.errors import PasswordDeleteError

PRIORITY = 10


def cli_exec(command: str):
    result = subprocess.run(command.split(" "), stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    return result.stdout, result.stderr


def _user_is_authenticated():
    command = "op whoami"
    stdout, stderr = cli_exec(command)
    if stderr:
        return False
    return True


def sign_in():
    command = "op signin"
    stdout, stderr = cli_exec(command)
    if stderr:
        raise RuntimeError("Failed to sign in to 1Password: " + stderr)
    return True


def _auth():
    if not _user_is_authenticated():
        sign_in()


def _item_exists(service, username):
    command = f"op item get {service} --fields username"
    stdout, stderr = cli_exec(command)
    if stderr:
        raise RuntimeError("Error checking if item exists: " + stderr)
    else:
        return stdout == username


def _onepassword_cli_installed():
    return bool(shutil.which("op"))


class OnePasswordBackend(backend.KeyringBackend):

    @properties.classproperty
    @classmethod
    def priority(cls):
        if not _onepassword_cli_installed():
            raise RuntimeError(
                "Requires onepassword cli: https://developer.1password.com/docs/cli/get-started/"
            )

        return PRIORITY

    def get_password(self, service, username):
        _auth()
        command = f"op item get {service} --fields username,password"
        stdout, stderr = cli_exec(command)
        if stderr:
            raise RuntimeError("Failed to get password: " + stderr)
        else:
            username, password = stdout.strip().split(",")
            if username != username:
                password = None
            return password.strip()

    def set_password(self, service, username, password):
        _auth()
        if _item_exists(service, username):
            command = f"op item edit {service} username={username} password={password}"
            stdout, stderr = cli_exec(command)
        else:
            command = f"op item create title={service} username={username} password={password} --category=login"
            stdout, stderr = cli_exec(command)
        if stderr:
            raise RuntimeError("Failed to set password: " + stderr)

    def delete_password(self, service, username):
        _auth()
        if _item_exists(service, username):
            command = f"op delete item {service}"
            stdout, stderr = cli_exec(command)
            if stderr:
                raise RuntimeError("Failed to delete password: " + stderr)
        else:
            e = sys.exc_info()[0]
            raise PasswordDeleteError(e)

    def get_otp(self, service, username):
        _auth()
        command = f"op item get {service} --otp"
        stdout, stderr = cli_exec(command)
        if stderr:
            raise RuntimeError("Failed to get otp: " + stderr)
        else:
            return stdout.strip()

    def get_item(self, service):
        _auth()
        command = f"op get item {service}"
        stdout, stderr = cli_exec(command)
        if stderr:
            raise RuntimeError("Failed to get item: " + stderr)
        return stdout.strip()
